    // the variable is defined
class VNPBridge {

    constructor() {

        if(this.constructor._tickerHandler == undefined) {

            this.constructor._tickerHandler = {}

        }

        if(this.constructor._eventBus == undefined) {

            this.constructor._eventBus = []

        }

        this.methods = new VNPMethods(this)
        this.i18n = new VNPVarContainer()
        this._uuid = ""
        this.args = {}
        this.config = {}
        this._version = ""

        this.ready = function() {
            const event = new Event('vnpAppReady');
            window.dispatchEvent(event);
        }

    }

    _getMobileOperatingSystem() {

        var userAgent = navigator.userAgent || navigator.vendor || window.opera;

        // Windows Phone must come first because its UA also contains "Android"
        if (/windows phone/i.test(userAgent)) {
            return "Windows Phone";
        }

        if (/android/i.test(userAgent)) {
            return "Android";
        }

        // iOS detection from: http://stackoverflow.com/a/9039885/177710
        if ((/iPad|iPhone|iPod/.test(userAgent) && !window.MSStream) || /Macintosh/.test(userAgent)) {
            return "iOS";
        }

        return "unknown";

    }

    _udid() {
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
            var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
            return v.toString(16);
        });
    }

    static resolvePromise(udid,data = "") {
        if(data != "") {
            data = decodeURIComponent(escape(atob(data)))
        }

        try {
            console.log(data)
            data = JSON.parse(data);
        } catch (e) {
            console.log("JSON EXCEPTION " + e)
        }
        console.log("TRY TO RESOLVE PROMISE " + udid + " - " + data)
        console.log(this.getPromise(udid))
        this.getPromise(udid).resolve(data)
    }

    static getPromise(udid) {

        return this._tickerHandler[udid]

    }

    static clearPromise(udid) {
        console.log("CLEAR " + udid)
        this._tickerHandler[udid] = null
        delete this._tickerHandler[udid]

    }

    addKeys() {

    }

    createPromise() {

        var _resolve, _reject;

        var promise = new Promise(function(resolve, reject){
            _resolve = resolve;
            _reject = reject;
        })

        promise.resolve = _resolve
        promise.reject = _reject

        let _ticketUdid = this._udid()

        console.log("CREATING PROMISE -> " + _ticketUdid)

        this.constructor._tickerHandler[_ticketUdid] = promise

        return _ticketUdid

    }

    subscribe(eventName,callback) {
        var obj = {}
        obj["callback"] = callback
        if(eventName != null) {
            eventName = eventName.toLowerCase()
        }
        obj["eventName"] = eventName
        this.constructor._eventBus.push(obj)
    }

    subscribeAll(callback) {

        var obj = {}
        obj["callback"] = callback
        obj["eventName"] = null
        this.constructor._eventBus.push(obj)

    }

    static emit(eventName, params) {
        var data = {}
        if(params != "") {
             data = atob(params)
         }
         try {
             data = JSON.parse(data);
         } catch (e) {
             console.log("JSON EXCEPTION " + e)
         }

        this._eventBus.forEach(function (obj) {

            if(obj["eventName"] == null || eventName.toLowerCase() == obj["eventName"]) {

                obj["callback"](eventName, data)

            }

        })

    }

    call(method, params = null) {

        var self = this;

        let ticketPromise = this.createPromise()

        var contract = {}
        contract["method"] = method
        contract["params"] = params
        contract["promise"] = ticketPromise

        let contractJson = JSON.stringify(contract);

        // to do defer
        window.setTimeout(function () {

            if(self._getMobileOperatingSystem() == "iOS") {

                webkit.messageHandlers.nativeProcess.postMessage(contractJson);

            } else if(self._getMobileOperatingSystem() == "Android") {

                android.nativeProcess(contractJson);

            }

        }, 1);

        return this.constructor.getPromise(ticketPromise)

    }

}

var VNPVarContainer = class VNPVarContainer {

    constructor() {

        this._data = {}

    }

    get(key) {

        var value = this._data[key]

        if(value == undefined) {
            return key
        } else {
            return value
        }

    }

}

class VNPMethods {

    constructor(bridge) {

        this.bridge = bridge

    }

    /*Device Stuff*/

    isiOS() {

        return this.bridge._getMobileOperatingSystem() == "iOS"

    }

    isAndroid() {
        return this.bridge._getMobileOperatingSystem() == "Android"
    }

    isOnline() {

        return this.bridge.call("isOnline");

    }

    /*Device Stuff*/

    /*User Staff*/

    isLogged() {

        return this.bridge.call("user.isLogged");

    }

    getUser() {

        return this.bridge.call("user.getUser");

    }

    setApiToken(token) {
        var params = {"token": token}
        return this.bridge.call("user.setApiToken", params);
    }

    getApiToken() {

        return this.bridge.call("user.getApiToken");

    }

    getVersion() {

        return this.bridge._version;

    }

    getUuid() {

        return this.bridge._uuid

    }

    logout() {

        return this.bridge.call("user.logout");

    }

    /*User Staff*/

    /*Versioning Staff*/

    setData(key, value) {
        var params = {"key": key, "value": value}
        return this.bridge.call("setData", params)
    }

    getData(key) {
        var params = {"key": key}
        return this.bridge.call("getData", params)
    }

    /*Versioning Staff*/

    getConfig(key) {
        var params = {"key": key}
        return this.bridge.call("getConfig", params)
    }

    showLoader() {
        this.bridge.call("showLoader")
    }

    hideLoader() {
        this.bridge.call("hideLoader")
    }

    http(url,method="GET",params = "", cache = true, failbackOffline = true, forceOfflineVersion = false) {
        var params = {"url": url, "method": method, "params": params, "cache" : cache, "failbackOffline" : failbackOffline, "forceOfflineVersion" : forceOfflineVersion}
        return this.bridge.call("http", params)
    }

    navigationControllerBack() {
        this.bridge.call("navigationControllerBack")
    }

    image(url,forceCache = false) {
        var params = {"url": url, "forceCache" : forceCache}
        return this.bridge.call("image", params)
    }

    showMenu() {
        return this.bridge.call("showMenu")
    }

    hideMenu() {
        return this.bridge.call("hideMenu")
    }

    location(url,name="self",topBar = true) {
        var params = {"url": url, "name": name, "topbar" : topBar}
        return this.bridge.call("location", params)
    }

    loadUrlApp(url,topBar = false) {
        var params = {"url": url, "topbar": topBar}
        return this.bridge.call("loadUrlApp", params)
    }

    loadLocalApp(source,force=false) {
        var params = {"source": source, "force": force}
        return this.bridge.call("loadLocalApp", params)
    }

    changeMainTab() {
        var params = {}
        return this.bridge.call("changeMainTab", params)
    }

    getPushToken() {
        var params = {}
        return this.bridge.call("getPushToken", params)
    }

    changeTab(id) {
        var params = {"id": id}
        return this.bridge.call("changeTab", params)
    }

    openEdition(id,page=0) {
        console.log("Opening Edition " + page)
        var params = {"id": id,"page": page}
        return this.bridge.call("openEdition", params)
    }

    getEditionArticle(editionId, articleId) {
        var params = {"article": articleId, "edition" : editionId}
        return this.bridge.call("getEditionArticle",params)
    }

    getEditionSummary(editionId) {
        var params = {"edition" : editionId}
        return this.bridge.call("getEditionSummary",params)
    }

    getPageFavorites() {
        return this.bridge.call("getFavorites")
    }

    getPriceProduct(product) {
        var params = {"product": product}
        return this.bridge.call("getPriceProduct",product)
    }

    deletePageFavorite(id) {
        var params = {"id": id}
        return this.bridge.call("deleteFavorite",params)
    }

    getDownloadedEditions() {
        return this.bridge.call("getDownloadedEditions")
    }

    deleteDownloadedEdition(id,correlate=false) {
        var params = {"id": id, "correlate" : correlate}
        return this.bridge.call("deleteDownloadedEdition", params)
    }

    closeActivity() {
        var params = {}
        return this.bridge.call("closeActivity", params)
    }

    closeOverlay() {
        var params = {}
        return this.bridge.call("closeOverlay", params)
    }

    purchaseEdition(productId,titleId = null,editionId = null) {
        var params = {"productId" : productId, "editionId" : editionId, "titleId" : titleId}
        return this.bridge.call("purchaseEdition", params)
    }

    purchaseEditionGooglePlay(productId,titleId = null,editionId = null) {
        var params = {"productId" : productId, "editionId" : editionId, "titleId" : titleId}
        return this.bridge.call("purchaseEditionGooglePlay", params)
    }

    getTitles() {
        return this.bridge.config.titles
    }

    getTitle(id) {
        var current_title = undefined;
        this.getTitles().forEach(function (title) {
            if(title.id == id) {
                current_title = title;
            }
        })
        return current_title
    }

    setLimitSavedEditions(limit) {
        var params = {"limit" : limit}
        return this.bridge.call("setLimitSavedEditions", params)
     }

    getLimitSavedEditions() {
        var params = {}
        return this.bridge.call("getLimitSavedEditions", params)
    }

    setCurrentTitle(id) {

        var self = this

        var title = this.getTitle(id)

        return new Promise(function(resolve,reject) {

            if(title != undefined) {

                self.setData("internal_current_title", id).then(result =>
                {

                    return resolve(title)

                });
            } else {
                return reject("Title not found")
            }

        });

    }

    getCurrentTitle() {

       var self = this

      return new Promise(function(resolve,reject) {

          self.getData("internal_current_title").then(result =>
          {

              var current_title = self.getTitle(result.value);

              if(current_title == undefined) {
                  current_title = self.getTitles()[0]
              }

              resolve(current_title)

          })

       })

    }

    alert(title,text,button = "OK",type = "sucess") {
        var params = {"title": title, "text": text, "type": type, "button": button}
        return this.bridge.call("alert", params)
    }

    share(title,link) {
        var params = {"title": title, "link": link}
        return this.bridge.call("share", params)
    }

    appEvent(type,bundle) {
        var params = {"type": type, "bundle": bundle}
        return this.bridge.call("appEvent", params)
    }

}

var vnpApp = new VNPBridge();

window.alert = function tmp(args) { vnpApp.methods.alert.apply(vnpApp.methods,arguments) };